package com.couchbase.test.utils;

import com.couchbase.client.java.Bucket;
import com.couchbase.client.java.Cluster;
import com.couchbase.client.java.CouchbaseCluster;
import com.couchbase.client.java.env.CouchbaseEnvironment;
import com.couchbase.client.java.env.DefaultCouchbaseEnvironment;
import com.couchbase.test.configurations.CouchbaseConfiguration;
import org.springframework.stereotype.Service;

/**
 * Couchbase Configuration Manager.
 */
@Service
public class CouchbaseConnectionManager {
  /**
   * Logger instance for the class.
   */
  private static final LoggingUtility logger = new LoggingUtility(CouchbaseStorageUtility.class);

  /**
   * Instance of {@link CouchbaseConfiguration}.
   */
  private final CouchbaseConfiguration couchbaseConfiguration;

  /**
   * The time period in milliseconds that a subscriber needs to subscribe to the observable.
   * After this period, the resources involved in the observable are released and can’t be
   * subscribed to anymore.
   */
  private static final int AUTO_RELEASE_AFTER = 10000;

  /**
   * Parameterized constructor.
   *
   * @param couchbaseConfiguration Auto injected instance of {@link CouchbaseConfiguration}.
   */
  public CouchbaseConnectionManager(CouchbaseConfiguration couchbaseConfiguration) {
    this.couchbaseConfiguration = couchbaseConfiguration;
  }

  /**
   * Creates a new connection to the Couchbase Cluster.
   *
   * @return Cluster Object.
   */
  public Cluster newClusterConnection() {
    CouchbaseEnvironment environment =
        DefaultCouchbaseEnvironment.builder().autoreleaseAfter(AUTO_RELEASE_AFTER).build();
    Cluster cluster = CouchbaseCluster.create(environment, this.couchbaseConfiguration.getUrl());
    cluster.authenticate(this.couchbaseConfiguration.getUsername(),
        this.couchbaseConfiguration.getPassword());
    logger.logInfo("newClusterConnection", "cluster information::" + cluster);
    return cluster;
  }

  /**
   * Creates a new connection to the bucket.
   *
   * @param cluster    Cluster configuration object.
   * @param bucketName Name of the bucket to create a connection.
   * @return Bucket Object.
   */
  public Bucket openBucket(Cluster cluster, String bucketName) {
    Bucket bucket = cluster.openBucket(bucketName);
    logger.logInfo("openBucket", "bucketInfo in connection manager::" + bucket);
    return bucket;
  }
}