package com.couchbase.test.utils;

import com.couchbase.client.java.Bucket;
import com.couchbase.client.java.Cluster;
import com.couchbase.client.java.document.AbstractDocument;
import com.couchbase.client.java.document.json.JsonObject;
import com.couchbase.client.java.query.AsyncN1qlQueryRow;
import com.couchbase.client.java.query.N1qlQuery;
import com.couchbase.client.java.query.ParameterizedN1qlQuery;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.couchbase.test.configurations.CouchbaseConfiguration;
import com.couchbase.test.constants.ErrorConstants;
import com.couchbase.test.exceptions.StorageException;

import java.util.HashMap;
import java.util.Map;

import org.reactivestreams.Publisher;
import org.springframework.stereotype.Repository;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;
import rx.Observable;
import rx.RxReactiveStreams;

/**
 * Contains methods for making Couchbase operations.
 */
@Repository
public class CouchbaseStorageUtility implements StorageUtility {

  /**
   * Logger instance for the class.
   */
  private static final LoggingUtility logger = new LoggingUtility(CouchbaseStorageUtility.class);

  /**
   * Auto injected instance of {@link CouchbaseConnectionManager}.
   */
  private CouchbaseConnectionManager couchbaseConnectionManager;

  /**
   * Auto injected instance of {@link CouchbaseConfiguration}.
   */
  private CouchbaseConfiguration couchbaseConfig;

  /**
   * Represents a singular cluster containing the buckets.
   */
  private Cluster cluster;

  /**
   * Maps buckets that will be handled by this application.
   */
  private Map<String, Bucket> buckets = new HashMap<>();

  /**
   * Message Constant.
   */
  private static final String MESSAGE = "msg";

  /**
   * Code constant.
   */
  private static final String CODE = "code";

  /**
   * Parameterized constructor.
   */
  public CouchbaseStorageUtility(CouchbaseConnectionManager couchbaseConnectionManager,
                                 CouchbaseConfiguration couchbaseConfiguration) {
    try {
      init(couchbaseConnectionManager, couchbaseConfiguration);
    } catch (Exception exception) {
      logger.logError("CouchbaseStorageUtilityImpl", "Couchbase not available.");
    }
  }

  /**
   * Retrieves the Couchbase document for the provided document id.
   */
  @Override
  public <T> Mono<T> get(String documentId, Class<T> documentClassType,
                         String bucketName) {
    logger.logEntry("get");
    Bucket bucket = this.buckets.get(bucketName);
    Observable<JsonObject> observable =
        bucket.async()
            .get(documentId)
            .map(AbstractDocument::content);
    Publisher<JsonObject> publisher = RxReactiveStreams.toPublisher(observable);
    return Mono.from(publisher)
        .flatMap(response -> this.mapResponse(response, documentClassType))
        .switchIfEmpty(Mono.error(new StorageException(
            ErrorConstants.DOCUMENT_DOES_NOT_EXIST.getErrorCode(),
            ErrorConstants.DOCUMENT_DOES_NOT_EXIST.getErrorMessage())));
  }

  /**
   * Executes ParameterizedN1qlQuery.
   * Overloaded method of execute query based on the query, params and bucket name.
   */
  public Flux<JsonObject> executeQuery(String query, String params, String bucketName) {
    logger.logEntry("executeQuery");
    //    CouchbaseEnvironment environment = DefaultCouchbaseEnvironment.builder()
    //        .autoreleaseAfter(10000).build();
    //    Cluster cluster = CouchbaseCluster.create(environment, this.couchbaseConfig.getUrl());
    //    cluster.authenticate(this.couchbaseConfig.getUsername(),
    //        this.couchbaseConfig.getPassword());
    // IF THE ABOVE CODE IS UNCOMMENTED, IT WORKS.

    logger.logInfo("executeQuery", "cluster information::" + cluster);
    Bucket bucket = cluster.openBucket(bucketName);

    logger.logInfo("executeQuery",
        "Couchbase bucket connection info execute query::" + bucket);
    ParameterizedN1qlQuery parameterizedN1qlQuery =
        N1qlQuery.parameterized(query, JsonObject.fromJson(params));
    Observable<JsonObject> observable = bucket.async()
        .query(parameterizedN1qlQuery)
        .flatMap(result -> result.errors()
            .flatMap(error -> Observable.<AsyncN1qlQueryRow>error(
                new StorageException(String.valueOf(
                    error.getInt(CODE)), error.getString(MESSAGE))))
            .switchIfEmpty(result.rows())
        )
        .map(AsyncN1qlQueryRow::value)
        .switchIfEmpty(Observable.just(JsonObject.empty()));
    Publisher<JsonObject> publisher = RxReactiveStreams.toPublisher(observable);
    return Flux.from(publisher);
  }

  /**
   * Executes ParameterizedN1qlQuery and maps the response.
   * Overloaded method of execute query based on the query, params, class type to be mapped,
   * and bucket name.
   */
  public <T> Flux<T> executeQuery(String query, String params,
                                  Class<T> documentClassType,
                                  String bucketName) {
    return this.executeQuery(query, params, bucketName)
        .flatMap(response -> mapResponse(response, documentClassType));
  }

  /**
   * Maps the Couchbase response to a POJO.
   *
   * @param jsonObject        Jsonobject to be mapped.
   * @param documentClassType Class type to map the object with.
   */
  private <T> Mono<T> mapResponse(JsonObject jsonObject, Class<T> documentClassType) {
    logger.logEntry("mapResponse");
    try {
      if (!jsonObject.isEmpty()) {
        return Mono.just(JsonUtility.mapToObject(jsonObject.toString(), documentClassType));
      }
      return Mono.empty();
    } catch (JsonProcessingException exception) {
      logger.logError("mapResponse", exception.getMessage());
      return Mono.error(new RuntimeException(ErrorConstants.JSON_PROCESSING_ERROR.getErrorCode()));
    }
  }

  /**
   * Manages Couchbase connections.
   *
   */
  private void init(CouchbaseConnectionManager couchbaseConnectionManager,
                    CouchbaseConfiguration couchbaseConfiguration) {
    logger.logEntry("init");
    this.couchbaseConnectionManager = couchbaseConnectionManager;
    this.cluster = couchbaseConnectionManager.newClusterConnection();
    logger.logInfo("init", "cluster information::" + this.cluster + " " + cluster + " ");
    this.couchbaseConfig = couchbaseConfiguration;
    this.couchbaseConfig.getBuckets().forEach((value) -> {
      Bucket bucket = this.couchbaseConnectionManager.openBucket(this.cluster, value);
      logger.logInfo("init", "Couchbase bucket connection info::" + bucket);
      buckets.put(value, bucket);
    });
  }
}
